import pytest
from superintendent import ClassLabeller
import superintendent.base
import superintendent.queueing
import superintendent.controls


def test_that_class_labeller_creates_base_widget():

    widget = ClassLabeller()

    assert isinstance(widget, superintendent.base.Labeller)


def test_that_class_labeller_has_expected_attributes():

    widget = ClassLabeller(options=("a", "b"))

    assert isinstance(
        widget.queue, superintendent.queueing.SimpleLabellingQueue
    )
    assert isinstance(widget.input_widget, superintendent.controls.Submitter)

    assert pytest.helpers.same_elements(
        widget.input_widget.options, ("a", "b")
    )


# from collections import namedtuple

# import numpy as np
# import pandas as pd
# import ipywidgets
# import ipyevents
# from sklearn.model_selection import cross_validate
# from sklearn.linear_model import LogisticRegression

# import pytest

# from hypothesis import given, settings
# from hypothesis.extra.numpy import (
#     array_shapes,
#     arrays,
#     boolean_dtypes,
#     floating_dtypes,
#     integer_dtypes,
#     unicode_string_dtypes,
# )
# from hypothesis.extra.pandas import column, columns, data_frames, series
# from hypothesis.strategies import (
#     booleans,
#     dictionaries,
#     floats,
#     integers,
#     lists,
#     one_of,
#     recursive,
#     text,
# )

# import superintendent.acquisition_functions
# from superintendent import ClassLabeller

# primitive_strategy = (
#   text() | integers() | floats(allow_nan=False) | booleans()
# )

# guaranteed_dtypes = (
#     boolean_dtypes()
#     | integer_dtypes()
#     | floating_dtypes()
#     | unicode_string_dtypes()
# )

# container_strategy = dictionaries(text(), primitive_strategy) | lists(
#     primitive_strategy
# )

# nested_strategy = recursive(
#     container_strategy,
#     lambda children: lists(children) | dictionaries(text(), children),
# )

# container_strategy = dictionaries(text(), primitive_strategy) | lists(
#     primitive_strategy
# )

# nested_strategy = recursive(
#     container_strategy,
#     lambda children: lists(children) | dictionaries(text(), children),
# )

# numpy_strategy = arrays(guaranteed_dtypes, array_shapes())

# pandas_series = series(dtype=int) | series(dtype=float) | series(dtype=str)

# pandas_dfs = (
#     data_frames(columns(3, dtype=int))
#     | data_frames(columns(3, dtype=float))
#     | data_frames(columns(3, dtype=str))
#     | data_frames([column(dtype=str),column(dtype=float),column(dtype=int)])
# )

# possible_input_data = one_of(
#     lists(primitive_strategy),
#     numpy_strategy,
#     pandas_series,
#     # pandas_dfs
# )


# TEST_DF = pd.DataFrame(np.meshgrid(np.arange(20), np.arange(20))[0])

# TEST_SERIES = pd.Series(np.arange(20))

# TEST_ARRAY = np.arange(20)

# TEST_LIST = list(range(20))

# TEST_LABELS_STR = {str(a) for a in np.arange(20)}
# TEST_LABELS_NUM = {float(a) for a in np.arange(20)}
# TEST_LABELS_CHAR = {"hello{}".format(a) for a in np.arange(20)}

# shuffle_opts = [True, False]

# data_opts = [TEST_DF, TEST_SERIES, TEST_ARRAY, TEST_LIST]

# label_opts = [TEST_LABELS_NUM, TEST_LABELS_STR, TEST_LABELS_CHAR]


# @pytest.mark.skip
# def test_that_creating_a_widget_works():
#     widget = ClassLabeller()  # noqa


# def test_that_supplied_features_are_passed_to_queue(mocker):
#     test_array = np.array([[1, 2, 3], [1, 2, 3]])
#     mock_queue = mocker.patch(
#         "superintendent.semisupervisor.SimpleLabellingQueue"
#     )
#     widget = ClassLabeller(features=test_array, labels=None)  # noqa
#     mock_queue.assert_called_once_with(test_array, None)


# def test_that_supplied_labels_are_passed_to_queue(mocker):
#     test_array = np.array([[1, 2, 3], [1, 2, 3]])
#     test_labels = np.array(["hi", "hello"])
#     mock_queue = mocker.patch(
#         "superintendent.semisupervisor.SimpleLabellingQueue"
#     )
#     widget = ClassLabeller(features=test_array, labels=test_labels)  # noqa
#     mock_queue.assert_called_once_with(test_array, test_labels)


# @given(shuffle_prop=floats(min_value=0, max_value=1))
# def test_that_shuffle_prop_is_set_correctly(shuffle_prop):
#     widget = ClassLabeller(shuffle_prop=shuffle_prop)
#     assert widget.shuffle_prop == shuffle_prop


# def test_that_classifier_is_set_correctly(mocker):
#     mock_classifier = mocker.Mock()
#     mock_classifier.fit == "dummy fit"
#     mock_classifier.predict_proba == "dummy predict"

#     mock_on_click = mocker.patch("ipywidgets.Button.on_click")

#     widget = ClassLabeller(classifier=mock_classifier)

#     assert widget.classifier is mock_classifier
#     assert hasattr(widget, "retrain_button")
#     assert isinstance(widget.retrain_button, ipywidgets.Button)
#     assert ((widget.retrain,),) in mock_on_click.call_args_list


# def test_that_eval_method_is_set_correctly(mocker):
#     mock_eval_method = mocker.Mock()

#     # test the default
#     widget = ClassLabeller()
#     assert widget.eval_method.func is cross_validate

#     # test the normal method
#     widget = ClassLabeller(eval_method=mock_eval_method)
#     assert widget.eval_method is mock_eval_method

#     # test the unhappy case
#     with pytest.raises(ValueError):
#         widget = ClassLabeller(eval_method="not a callable")


# def test_that_reorder_is_set_correctly(mocker):
#     mock_reorder_method = mocker.Mock()

#     # test the default
#     widget = ClassLabeller()
#     assert widget.reorder is None

#     # test passing a string
#     widget = ClassLabeller(reorder="entropy")
#     assert widget.reorder is superintendent.acquisition_functions.entropy

#     # test a function
#     widget = ClassLabeller(reorder=mock_reorder_method)
#     assert widget.reorder is mock_reorder_method

#     # test the unhappy case
#     with pytest.raises(NotImplementedError):
#         widget = ClassLabeller(reorder="dummy function name")

#     with pytest.raises(ValueError):
#         widget = ClassLabeller(reorder=1)


# def test_that_the_control_widget_calls_apply_annotation(mocker):
#     mock_submit = mocker.patch(
#         "superintendent.queueing.SimpleLabellingQueue.submit"
#     )
#     test_array = np.array([[1, 2, 3], [1, 2, 3]])

#     widget = ClassLabeller(features=test_array, options=["dummy label"])
#     widget.input_widget._when_submitted(
#         widget.input_widget.control_elements.buttons["dummy label"]
#     )

#     mock_submit.assert_called_once_with(0, "dummy label")


# def test_that_sending_labels_into_iterator_submits_them_to_queue(mocker):
#     mock_submit = mocker.patch(
#         "superintendent.queueing.SimpleLabellingQueue.submit"
#     )
#     test_array = np.array([[1, 2, 3], [1, 2, 3]])

#     widget = ClassLabeller(features=test_array)
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})

#     mock_submit.assert_called_once_with(0, "dummy label")


# def test_that_sending_undo_into_iterator_calls_undo_on_queue(mocker):
#     mock_undo = mocker.patch(
#         "superintendent.queueing.SimpleLabellingQueue.undo"
#     )

#     test_array = np.array([[1, 2, 3], [1, 2, 3]])
#     widget = ClassLabeller(features=test_array)
#     widget._annotation_loop.send({"source": "__undo__", "value": None})

#     assert mock_undo.call_count == 2


# def test_that_sending_skip_calls_no_queue_method(mocker):
#     mock_undo = mocker.patch(
#         "superintendent.queueing.SimpleLabellingQueue.undo"
#     )
#     mock_submit = mocker.patch(
#         "superintendent.queueing.SimpleLabellingQueue.submit"
#     )

#     test_array = np.array([[1, 2, 3], [1, 2, 3]])
#     widget = ClassLabeller(features=test_array)
#     widget._annotation_loop.send({"source": "__skip__", "value": None})

#     assert mock_undo.call_count == 0
#     assert mock_submit.call_count == 0


# def test_that_added_labels_are_returned_correctly(mocker):

#     test_array = np.array([[1, 2, 3], [1, 2, 3]])
#     widget = ClassLabeller(features=test_array)

#     widget._annotation_loop.send({"source": "", "value": "dummy label"})
#     widget._annotation_loop.send({"source": "", "value": "dummy label 2"})

#     assert widget.new_labels == ["dummy label", "dummy label 2"]


# def test_that_progressbar_value_is_updated_and_render_finished_called(
#   mocker
# ):
#     mock_render_finished = mocker.patch(
#         "superintendent.semisupervisor.SemiSupervisor._render_finished"
#     )

#     test_array = np.array([[1, 2, 3], [1, 2, 3]])
#     widget = ClassLabeller(features=test_array)

#     assert widget.progressbar.value == 0
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})
#     assert widget.progressbar.value == 0.5
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})
#     assert widget.progressbar.value == 1
#     assert mock_render_finished.call_count == 1


# def test_that_calling_retrain_without_classifier_breaks():
#     with pytest.raises(ValueError):
#         test_array = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
#         widget = ClassLabeller(features=test_array)
#         widget.retrain()


# def test_that_retrain_with_no_labels_sets_warnings(mocker):
#     mock_classifier = mocker.Mock()
#     mock_classifier.fit == mocker.Mock()
#     mock_classifier.predict_proba == mocker.Mock()

#     mock_eval_method = mocker.Mock(
#         return_value={"test_score": np.array([0.8])}
#     )

#     test_array = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
#     widget = ClassLabeller(
#         features=test_array,
#         classifier=mock_classifier,
#         eval_method=mock_eval_method,
#     )

#     widget.retrain()
#     assert (
#         widget.model_performance.value
#         == "Score: Not enough labels to retrain."
#     )

#     widget._annotation_loop.send({"source": "", "value": "dummy label 1"})

#     widget.retrain()
#     assert (
#         widget.model_performance.value
#         == "Score: Not enough labels to retrain."
#     )

#     widget._annotation_loop.send({"source": "", "value": "dummy label 2"})

#     widget.retrain()

#     assert mock_eval_method.call_count == 1
#     assert widget.model_performance.value == "Score: 0.80"


# def test_that_retrain_calls_eval_method_correctly(mocker):
#     mock_classifier = mocker.Mock()
#     mock_classifier.fit == mocker.Mock()
#     mock_classifier.predict_proba == mocker.Mock()

#     mock_eval_method = mocker.Mock(
#         return_value={"test_score": np.array([0.8])}
#     )

#     test_array = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
#     widget = ClassLabeller(
#         features=test_array,
#         classifier=mock_classifier,
#         eval_method=mock_eval_method,
#     )

#     widget._annotation_loop.send({"source": "", "value": "dummy label 1"})
#     widget._annotation_loop.send({"source": "", "value": "dummy label 2"})
#     widget.retrain()

#     assert mock_eval_method.call_count == 1

#     call_arguments = mock_eval_method.call_args[0]
#     assert call_arguments[0] is mock_classifier
#     assert (call_arguments[1] == test_array[:2, :]).all()
#     assert pytest.helpers.same_elements(
#         call_arguments[2], ["dummy label 1", "dummy label 2"]
#     )
#     assert widget.model_performance.value == "Score: 0.80"


# def test_that_retrain_calls_reorder_correctly(mocker):

#     test_probabilities = np.array([[0.2, 0.3], [0.1, 0.4]])

#     mock_eval_method = mocker.Mock(
#         return_value={"test_score": np.array([0.8])}
#     )

#     mock_reordering = mocker.Mock(return_value=[0, 1])

#     test_array = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3], [1, 2, 3]])

#     widget = ClassLabeller(
#         features=test_array,
#         classifier=LogisticRegression(),
#         eval_method=mock_eval_method,
#         reorder=mock_reordering,
#         shuffle_prop=0.2,
#     )

#     mocker.patch.object(
#         widget.classifier, "fit", return_value=LogisticRegression()
#     )
#     mocker.patch.object(
#         widget.classifier, "predict_proba", return_value=test_probabilities
#     )

#     widget._annotation_loop.send({"source": "", "value": "dummy label 1"})
#     widget._annotation_loop.send({"source": "", "value": "dummy label 2"})
#     widget.retrain()

#     assert mock_reordering.call_count == 1

#     call_args, call_kwargs = mock_reordering.call_args_list[0]

#     assert (call_args[0] == test_probabilities).all()
#     assert call_kwargs["shuffle_prop"] == 0.2


# def test_that_the_event_manager_is_closed(mocker):

#     test_array = np.array([[1, 2, 3], [1, 2, 3]])

#     mock_event_manager_close = mocker.patch.object(ipyevents.Event, "close")

#     widget = ClassLabeller(features=test_array, keyboard_shortcuts=True)
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})

#     assert mock_event_manager_close.call_count == 1



"""Functions to validate arguments."""
# from typing import Any


# def valid_classifier(classifier: Any):
#     """
#     Check if an object conforms to sklearns fit / predict interface.

#     Parameters
#     ----------
#     classifier : sklearn.base.ClassifierMixin
#         A classification model compliant with sklearn interfaces.
#     """
#     if classifier is None:
#         return None
#     elif hasattr(classifier, "fit") and hasattr(classifier, "predict_proba"):
#         return classifier
#     else:
#         raise ValueError(
#             "The classifier needs to conform to "
#             "the sklearn interface (fit/predict_proba)."
#         )


# def valid_data(features: Optional[Any]):
#     """
#     Check if an object is an array or can be turned into one.

#     Parameters
#     ----------
#     features : pd.DataFrame, pd.Series, np.ndarray
#         the data to double-check.
#     """
#     if features is None:
#         return None
#     if isinstance(
#         features, (pd.DataFrame, pd.Series, np.ndarray, list, tuple)
#     ):
#         return features
#     # elif isinstance(features, (list, tuple)):
#     #     return np.array(features)
#     else:
#         raise ValueError(
#             "The features need to be an array, sequence, or "
#             "a pandas DataFrame / Series."
#         )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['superintendent',
 'superintendent.acquisition_functions',
 'superintendent.controls',
 'superintendent.display',
 'superintendent.distributed',
 'superintendent.queueing']

package_data = \
{'': ['*']}

install_requires = \
['ipywidgets>=7.4',
 'numpy>=1.15',
 'pandas>=0.24.0',
 'scikit-learn>=0.20.3',
 'scipy>=1.2',
 'schedule>=0.6.0',
 'sqlalchemy>=1.3',
 'Pillow',
 'cachetools>=3.1',
 'psycopg2-binary>=2.8',
 'flask>=1.0',
 'ipyevents>=0.6.0']

extras_require = \
{'dev': ['jupyterlab>=1.1', 'rope', 'docargs', 'black>=19.3b0', 'flake8>=3.7'],
 'doc': ['sphinx>=2.0',
         'requests>=2.21',
         'html5lib>=1.0',
         'bs4>=0.0.1',
         'wordcloud>=1.5',
         'm2r>=0.2.1',
         'sphinx_rtd_theme>=0.4.3',
         'jupyter-sphinx>=0.1.4',
         'nbsphinx>=0.4.2'],
 'examples': ['requests>=2.21',
              'html5lib>=1.0',
              'bs4>=0.0.1',
              'wordcloud>=1.5'],
 'test': ['isort>=4.3.21',
          'pytest>=4.1',
          'pytest-cov>=2.6',
          'hypothesis>=4.17',
          'pytest-helpers-namespace>=2019.1',
          'pytest-mock>=1.10',
          'nbconvert>=5.4',
          'coveralls>=1.7']}

setup(name='superintendent',
      version='0.5.0rc1',
      description='Interactive machine learning supervision.',
      author='Jan Freyberg',
      author_email='jan.freyberg@gmail.com',
      url='https://github.com/janfreyberg/superintendent',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



import time

from sklearn.datasets import load_digits
from sklearn.linear_model import LogisticRegression
from sqlalchemy.exc import OperationalError
from superintendent.distributed import SemiSupervisor

model = LogisticRegression(solver="lbfgs", max_iter=2000, multi_class="auto")

db_connection = (
    "postgresql+psycopg2://superintendent:superintendent" "@db:5432/labelling"
)

time.sleep(15)
connection_made = False
tries = 0

# wait for our database to come up:
while not connection_made and tries < 25:
    try:
        widget = SemiSupervisor.from_images(
            connection_string=db_connection,
            options=range(10),
            classifier=model,
            reorder="entropy",
            shuffle_prop=0.2,
        )
        connection_made = True
    except OperationalError:
        time.sleep(2)
        tries += 1

# if we've never added any data to this db, load it and add it:
if len(widget.queue) == 0:
    digit_data = load_digits().data
    widget.add_features(digit_data)

widget.orchestrate(interval_seconds=30, interval_n_labels=10)



import pytest

pytestmark = pytest.mark.skip

# import numpy as np
# import pandas as pd

# from collections import OrderedDict

# from hypothesis import given, settings
# from hypothesis.strategies import (
#     booleans,
#     floats,
#     integers,
#     lists,
#     tuples,
#     one_of,
#     sampled_from,
#     text,
#     composite,
# )
# from hypothesis.extra.pandas import data_frames, column
# from hypothesis.extra.numpy import (
#     arrays,
#     scalar_dtypes,
#     unsigned_integer_dtypes,
#     datetime64_dtypes,
#     floating_dtypes,
#     integer_dtypes,
# )

# from hypothesis import HealthCheck

# from superintendent.queueing import ClusterLabellingQueue


# guaranteed_dtypes = one_of(
#     scalar_dtypes(),
#     unsigned_integer_dtypes(),
#     datetime64_dtypes(),
#     floating_dtypes(),
#     integer_dtypes(),
# )


# @composite
# def dataframe(draw, length=None):
#     n_cols = draw(integers(min_value=1, max_value=20))
#     dtypes = draw(
#         lists(
#             sampled_from([float, int, str]), min_size=n_cols, max_size=n_cols
#         )
#     )
#     colnames = draw(
#         lists(
#             text() | integers(), min_size=n_cols, max_size=n_cols,
#             unique=True
#         )
#     )
#     df = draw(
#         data_frames(
#             columns=[
#                 column(name=name, dtype=dtype)
#                 for dtype, name in zip(dtypes, colnames)
#             ]
#         )
#     )
#     return df


# @composite
# def dataframe_and_clusters(draw, length=None):
#     n_cols = draw(integers(min_value=1, max_value=20))
#     dtypes = draw(
#         lists(
#             sampled_from([float, int, str]), min_size=n_cols, max_size=n_cols
#         )
#     )
#     colnames = draw(
#         lists(
#             text() | integers(), min_size=n_cols, max_size=n_cols,
#             unique=True
#         )
#     )
#     df = draw(
#         data_frames(
#             columns=[
#                 column(name=name, dtype=dtype)
#                 for dtype, name in zip(dtypes, colnames)
#             ]
#         )
#     )
#     cluster_labels = draw(
#         lists(
#             integers(min_value=0, max_value=3),
#             min_size=len(df),
#             max_size=len(df),
#         )
#     )
#     return df, cluster_labels


# @composite
# def array_and_clusters(draw):
#     array = draw(
#         arrays(
#             guaranteed_dtypes,
#             tuples(
#                 integers(min_value=1, max_value=50),
#                 integers(min_value=1, max_value=50),
#             ),
#         )
#     )
#     cluster_labels = draw(
#         lists(
#             integers(min_value=0, max_value=3),
#             min_size=array.shape[0],
#             max_size=array.shape[0],
#         )
#     )
#     return array, cluster_labels


# @given(
#     input_=one_of(booleans(), floats(), integers(), text()),
#     cluster_index=integers(),
# )
# def test_enqueueing_and_popping(input_, cluster_index):
#     q = ClusterLabellingQueue()

#     for i in range(10):
#         q.enqueue(cluster_index, input_)

#     idx, data = q.pop()
#     assert idx == cluster_index
#     assert pytest.helpers.same_elements(data, [input_] * 10)
#     assert pytest.helpers.same_elements(data, [input_] * 10)


# @given(
#     inputs=lists(one_of(booleans(), floats(), integers(), text()),
# min_size=1),
#     cluster_index=integers(),
# )
# def test_enqueue_many(inputs, cluster_index):
#     q = ClusterLabellingQueue()
#     q.enqueue_many(inputs, [cluster_index] * len(inputs))
#     assert len(q.data) == 1
#     # assert we can pop the one cluster:
#     q.pop()
#     # assert there's nothing else to pop:
#     with pytest.raises(IndexError):
#         q.pop()


# @settings(suppress_health_check=(HealthCheck.too_slow,))
# @given(inputs=dataframe_and_clusters())
# def test_enqueue_dataframe(inputs):

#     inputs, cluster_labels = inputs

#     n = len(inputs)
#     q = ClusterLabellingQueue()
#     q.enqueue_many(inputs, cluster_labels)

#     # assert the queue has only unique elements from cluster_labels
#     assert pytest.helpers.same_elements(q.order, set(cluster_labels))

#     # assert len(q.data) == n
#     # # assert we can pop everything:
#     for _ in range(len(set(cluster_labels))):
#         id_, val = q.pop()
#         assert isinstance(val, pd.DataFrame)
#     # assert there's nothing else to pop:
#     with pytest.raises(IndexError):
#         q.pop()
#     # assert it re-constructs a df on list all
#     if n > 0:
#         ids, X, y = q.list_all()
#         assert isinstance(X, pd.DataFrame) or len(X) == 0
#         # assert it re-constructs a df on list uncomplete
#         q.submit(cluster_labels[0], "hello")
#         ids, X = q.list_uncompleted()
#         assert isinstance(X, pd.DataFrame) or len(X) == 0
#         # assert it re-constructs a df on list uncomplete
#         ids, X, y = q.list_completed()
#         assert isinstance(X, pd.DataFrame) or len(X) == 0


# @settings(suppress_health_check=(HealthCheck.too_slow,))
# @given(inputs=array_and_clusters())
# def test_enqueue_array(inputs):

#     inputs, cluster_labels = inputs

#     n_features = inputs.shape[0]
#     n_clusters = len(set(cluster_labels))
#     q = ClusterLabellingQueue()
#     q.enqueue_many(inputs, cluster_labels)

#     assert len(q.data) == n_clusters
#     # assert we can pop everything:
#     for _ in range(n_clusters):
#         id_, val = q.pop()
#         assert isinstance(val, np.ndarray)
#         assert len(val.shape) == len(inputs.shape)
#         assert val.shape[-1] == inputs.shape[-1]

#     # assert there's nothing else to pop:
#     with pytest.raises(IndexError):
#         q.pop()
#     # assert it re-constructs an df on list all
#     if n_clusters > 0 and n_features > 0:
#         ids, X, y = q.list_all()
#         assert isinstance(X, np.ndarray)
#         # assert it re-constructs a df on list uncomplete
#         ids, X = q.list_uncompleted()
#         assert isinstance(X, np.ndarray)

#         # assert it re-constructs a df on list complete
#         q.submit(cluster_labels[0], "hello")
#         ids, X, y = q.list_completed()
#         assert isinstance(X, np.ndarray)


# @given(inputs=array_and_clusters())
# def test_enqueue_at_creation(inputs):

#     inputs, cluster_labels = inputs

#     n_clusters = len(set(cluster_labels))

#     q = ClusterLabellingQueue(inputs, cluster_labels)

#     assert len(q.data) == n_clusters
#     # assert we can pop everything:
#     for _ in range(n_clusters):
#         q.pop()
#     # assert there's nothing else to pop:
#     with pytest.raises(IndexError):
#         q.pop()


# @given(label1=text(), label2=text())
# def test_submitting_text(label1, label2):
#     q = ClusterLabellingQueue()
#     q.enqueue(0, 1)
#     q.enqueue(0, 1.5)
#     q.enqueue(1, 1)
#     q.enqueue(1, 1.5)
#     # assert you can't submit before popping:
#     with pytest.raises(ValueError):
#         q.submit(0, label1)
#     # pop the cluster and then submit a label:
#     id_, val = q.pop()
#     q.submit(id_, label1)
#     id_, val = q.pop()
#     q.submit(id_, label2)
#     assert q.list_labels() == {label1, label2}


# @given(label1=text(), label2=text())
# def test_submitting_list(label1, label2):
#     q = ClusterLabellingQueue()
#     q.enqueue(0, 1)
#     with pytest.raises(ValueError):
#         q.submit(0, label1)
#     id_, val = q.pop()
#     q.submit(id_, [label1, label2])
#     assert q.list_labels() == {label1, label2}


# @pytest.mark.skip  # skipping because this isn't implemented yet
# def test_reordering():
#     inp = ["b", "a", "d", "c"]
#     cluster_labels = [0, 1, 2, 3]
#     q = ClusterLabellingQueue()
#     q.enqueue_many(inp, cluster_labels)
#     q.reorder(OrderedDict([(0, 1), (1, 0), (2, 3), (3, 2)]))

#     id_, val = q.pop()
#     assert val == ["a"] and id_ == 1
#     id_, val = q.pop()
#     assert val == "b" and id_ == 0
#     id_, val = q.pop()
#     assert val == "c" and id_ == 3
#     id_, val = q.pop()
#     assert val == "d" and id_ == 2


# def test_iterating_over_queue():
#     inps = [str(i) for i in range(50)]
#     cluster_labels = [np.random.randint(0, 5) for i in range(50)]
#     n_clusters = len(set(cluster_labels))
#     q = ClusterLabellingQueue()
#     q.enqueue_many(inps, cluster_labels)

#     for i, (cluster_index, cluster_data) in enumerate(q):
#         pass
#     assert i + 1 == n_clusters


# def test_length_of_queue():
#     inps = [str(i) for i in range(50)]
#     cluster_labels = [np.random.randint(0, 5) for i in range(50)]
#     n_clusters = len(set(cluster_labels))
#     q = ClusterLabellingQueue()
#     assert len(q) == 0
#     q.enqueue_many(inps, cluster_labels)
#     assert len(q) == n_clusters


# def test_progress():
#     inps = [str(i) for i in range(50)]
#     cluster_labels = [np.random.randint(0, 5) for i in range(50)]
#     n_clusters = len(set(cluster_labels))
#     q = ClusterLabellingQueue()
#     # ensure zero divisions are caught
#     assert np.isnan(q.progress)
#     # ensure progress increases proportionally
#     q.enqueue_many(inps, cluster_labels)
#     for i, (id_, val) in enumerate(q):
#         assert q.progress == i / n_clusters
#         q.submit(id_, str(i))


# def test_shuffling():
#     inps = [str(i) for i in range(50)]
#     cluster_labels = list(range(50))
#     q = ClusterLabellingQueue()
#     q.enqueue_many(inps, cluster_labels)
#     q.shuffle()
#     # assert the order is not the same:
#     assert not all([val == [inp] for inp, (id_, val) in zip(inps, q)])


# def test_undo():
#     inp = "input 1"
#     cluster_index = 0
#     q = ClusterLabellingQueue()
#     q.enqueue(cluster_index, inp)
#     id_, val = q.pop()
#     q.submit(id_, "label 1")
#     # ensure the queue is empty now:
#     with pytest.raises(IndexError):
#         q.pop()
#     q.undo()
#     # see if it's possible to pop now:
#     id_, val = q.pop()
#     assert val == ["input 1"]


# @given(
#     inputs=lists(one_of(booleans(), floats(), integers(), text()),
# min_size=5),
#     labels=lists(text(), min_size=5),
# )
# def test_list_completed(inputs, labels):
#     q = ClusterLabellingQueue()
#     cluster_labels = [np.random.randint(0, 5) for _ in inputs]
#     n_clusters = len(set(cluster_labels))
#     q.enqueue_many(inputs, cluster_labels)

#     popped_ids = []
#     for i in range(n_clusters // 2):
#         id_, val = q.pop()
#         q.submit(id_, labels[i])

#         popped_ids.append(id_)

#     cluster_indices, x, y = q.list_completed()

#     assert len(set(cluster_indices)) == n_clusters // 2
#     # test that the popped IDs and completed IDs have the same members
#     assert pytest.helpers.same_elements(set(cluster_indices), popped_ids)
#     assert pytest.helpers.same_elements(
# set(y), set(labels[: n_clusters // 2]))


# @given(
#     inputs=lists(one_of(booleans(), floats(), integers(), text()),
# min_size=5),
#     labels=lists(text(), min_size=5),
# )
# def test_list_uncompleted(inputs, labels):
#     q = ClusterLabellingQueue()
#     cluster_labels = [np.random.randint(0, 5) for _ in inputs]
#     n_clusters = len(set(cluster_labels))
#     q.enqueue_many(inputs, cluster_labels)

#     popped_ids = []
#     for i in range(n_clusters // 2):
#         id_, val = q.pop()
#         q.submit(id_, labels[i])

#         popped_ids.append(id_)

#     cluster_indices, x = q.list_uncompleted()

#     assert len(set(cluster_indices)) == n_clusters - n_clusters // 2
#     # test that the popped IDs and completed IDs don't share members
#     assert pytest.helpers.no_shared_members(cluster_indices, popped_ids)


# def test_list_all():

#     inps = [str(i) for i in range(50)]
#     cluster_labels = [np.random.randint(0, 5) for i in range(50)]
#     n_clusters = len(set(cluster_labels))
#     labels = [str(np.random.rand()) for i in range(n_clusters)]

#     q = ClusterLabellingQueue()
#     q.enqueue_many(inps, cluster_labels)

#     popped_ids = []
#     for i in range(n_clusters // 2):
#         id_, val = q.pop()
#         q.submit(id_, labels[i])

#         popped_ids.append(id_)

#     cluster_indices, x, y = q.list_all()

#     assert len(cluster_indices) == len(inps)
#     # ensure not the labels have been listed:
#     assert not all(label is None for label in y)
#     assert all(
#         [
#             label is None or id_ in popped_ids
#             for id_, label in zip(cluster_indices, y)
#         ]
#     )
#     assert len(set(y) - {None}) == n_clusters // 2
#     assert pytest.helpers.same_elements(cluster_indices, cluster_labels)



from collections import defaultdict

import ipywidgets as widgets
import traitlets


class HintedMultiselect(widgets.HBox):

    options = traitlets.List(list())
    value = traitlets.List(list())

    def __init__(self, options, *args, **kwargs):
        """Created a Multi-select widget

        Parameters
        ----------
        options : Sequence[str]
            The options to show in the multi-select widget.
        """

        super().__init__([])

        self.options = [str(option) for option in options]
        self.multi_select = widgets.SelectMultiple(
            options=self.options, description="Label:"
        )
        widgets.link((self, "options"), (self.multi_select, "options"))
        widgets.link((self, "value"), (self.multi_select, "value"))

        self.hints = defaultdict(widgets.Output)

        self.children = [
            self.multi_select,
            widgets.HBox(
                children=[self.hints[option] for option in self.value],
                layout=widgets.Layout(flex_flow="row wrap"),
            ),
        ]

    def _reset(self):
        self.value = []

    def _toggle(self, option: str):

        if option in self.value:
            new_value = list(self.value)
            new_value.remove(option)
            self.multi_select.value = new_value
        else:
            self.multi_select.value = self.value + [option]

    @traitlets.observe("value")
    def _refresh_hints(self, change=None):
        self.children = [
            self.multi_select,
            widgets.HBox(
                children=[self.hints[option] for option in self.value],
                layout=widgets.Layout(flex_flow="row wrap"),
            ),
        ]
